

set -ex



test -L $PREFIX/lib/libcurand.so
test -f $PREFIX/targets/sbsa-linux/include/curand.h
test -L $PREFIX/targets/sbsa-linux/lib/libcurand.so
test -f $PREFIX/targets/sbsa-linux/lib/stubs/libcurand.so
exit 0
